#pragma once

// Low-level

#include <jni/types.hpp>
#include <jni/functions.hpp>
#include <jni/npe.hpp>


// High-level

#include <jni/unique.hpp>
#include <jni/tagging.hpp>
#include <jni/class.hpp>
#include <jni/object.hpp>
#include <jni/string.hpp>
#include <jni/array.hpp>
#include <jni/constructor.hpp>
#include <jni/method.hpp>
#include <jni/static_method.hpp>
#include <jni/field.hpp>
#include <jni/static_field.hpp>
#include <jni/native_method.hpp>
#include <jni/boxing.hpp>
#include <jni/advanced_ownership.hpp>
#include <jni/weak_reference.hpp>
